//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Core/Scan/CoordSystem1D.cpp
//! @brief     Implements CoordSystem1D class and derived classes.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Core/Scan/CoordSystem1D.h"
#include "Base/Axis/PointwiseAxis.h"
#include "Base/Const/Units.h"
#include "Base/Math/Constants.h"
#include "Core/Scan/AngularSpecScan.h"
#include "Core/Scan/QSpecScan.h"
#include "Device/Beam/Beam.h"
#include "Device/Data/OutputData.h"
#include "Device/Unit/AxisNames.h"

namespace {

double getQ(double wavelength, double angle)
{
    return 4.0 * M_PI * std::sin(angle) / wavelength;
}

double getInvQ(double wavelength, double q)
{
    double sin_angle = q * wavelength / (4.0 * M_PI);
    return std::asin(sin_angle);
}

std::unique_ptr<PointwiseAxis>
createTranslatedAxis(const IAxis& axis, std::function<double(double)> translator, std::string name)
{
    auto coordinates = axis.binCenters();
    for (size_t i = 0, size = coordinates.size(); i < size; ++i)
        coordinates[i] = translator(coordinates[i]);
    return std::make_unique<PointwiseAxis>(name, coordinates);
}

} // namespace


//  ************************************************************************************************
//  class CoordSystem1D
//  ************************************************************************************************

std::unique_ptr<CoordSystem1D> CoordSystem1D::createCoordSystem(const ISpecularScan& scan)
{
    if (const auto* aScan = dynamic_cast<const AngularSpecScan*>(&scan))
        return std::make_unique<AngularReflectometryConverter>(*aScan);

    if (const auto* qScan = dynamic_cast<const QSpecScan*>(&scan))
        return std::make_unique<WavenumberReflectometryConverter>(*qScan);

    throw std::runtime_error("Bug in CoordSystem1D::createCoordSystem: invalid case");
}

//! Returns the size of underlying axis.
size_t CoordSystem1D::axisSize(size_t i_axis) const
{
    ASSERT(i_axis == 0);
    return coordinateAxis()->size();
}

double CoordSystem1D::calculateMin(size_t i_axis, Axes::Units units) const
{
    ASSERT(i_axis == 0);
    units = substituteDefaultUnits(units);
    if (units == Axes::Units::NBINS)
        return 0.0;
    auto translator = getTraslatorTo(units);
    return translator(coordinateAxis()->binCenter(0));
}

double CoordSystem1D::calculateMax(size_t i_axis, Axes::Units units) const
{
    ASSERT(i_axis == 0);
    units = substituteDefaultUnits(units);
    const IAxis *const coordinate_axis = coordinateAxis();
    if (units == Axes::Units::NBINS)
        return static_cast<double>(coordinate_axis->size());
    auto translator = getTraslatorTo(units);
    return translator(coordinate_axis->binCenter(coordinate_axis->size() - 1));
}

std::unique_ptr<const IAxis> CoordSystem1D::createConvertedAxis(size_t i_axis,
                                                                  Axes::Units units) const
{
    ASSERT(i_axis == 0);
    units = substituteDefaultUnits(units);
    if (units == Axes::Units::NBINS)
        return std::make_unique<FixedBinAxis>(axisName(0, units), coordinateAxis()->size(),
                                              calculateMin(0, units), calculateMax(0, units));
    return createTranslatedAxis(*coordinateAxis(), getTraslatorTo(units), axisName(0, units));
}

std::unique_ptr<OutputData<double>>
CoordSystem1D::createConvertedData(const OutputData<double>& data, Axes::Units units) const
{
    ASSERT(data.rank() == 1);

    std::unique_ptr<OutputData<double>> result(new OutputData<double>);
    auto q_axis = createConvertedAxis(0, units);
    result->addAxis(*q_axis);

    if (units != Axes::Units::RQ4) {
        result->setRawDataVector(data.getRawDataVector());
        return result;
    }

    for (size_t i = 0, size = result->getAllocatedSize(); i < size; ++i)
        (*result)[i] = data[i] * std::pow((*q_axis)[i], 4);
    return result;
}


//  ************************************************************************************************
//  class AngularReflectometryConverter
//  ************************************************************************************************

AngularReflectometryConverter::AngularReflectometryConverter(const Beam& beam, const IAxis& axis,
                                                             Axes::Units axis_units)
    : m_wavelength(beam.wavelength())
    , m_axis(createTranslatedAxis(axis, getTraslatorFrom(axis_units), axisName(0, axis_units)))

{
    if (m_axis->lowerBound() < 0 || m_axis->upperBound() > M_PI_2)
        throw std::runtime_error("Error in CoordSystem1D: input axis range is out of bounds");
}

AngularReflectometryConverter::AngularReflectometryConverter(const AngularSpecScan& handler)
    : m_wavelength(handler.wavelength()), m_axis(handler.coordinateAxis()->clone())
{
}

AngularReflectometryConverter::~AngularReflectometryConverter() = default;

AngularReflectometryConverter* AngularReflectometryConverter::clone() const
{
    return new AngularReflectometryConverter(*this);
}

std::vector<Axes::Units> AngularReflectometryConverter::availableUnits() const
{
    return {Axes::Units::NBINS, Axes::Units::RADIANS, Axes::Units::DEGREES, Axes::Units::QSPACE,
            Axes::Units::RQ4};
}

AngularReflectometryConverter::AngularReflectometryConverter(
    const AngularReflectometryConverter& other)
    : m_wavelength(other.m_wavelength), m_axis(other.m_axis->clone())
{
}

std::vector<std::map<Axes::Units, std::string>>
AngularReflectometryConverter::createNameMaps() const
{
    return {AxisNames::specAxis};
}

std::function<double(double)>
AngularReflectometryConverter::getTraslatorFrom(Axes::Units units) const
{
    switch (units) {
    case Axes::Units::RADIANS:
        return [](double value) { return value; };
    case Axes::Units::DEGREES:
        return [](double value) { return Units::deg2rad(value); };
    case Axes::Units::QSPACE:
        return [this](double value) { return getInvQ(m_wavelength, value); };
    default:
        throwUnitsError("AngularReflectometryConverter::getTraslatorFrom",
                        {Axes::Units::RADIANS, Axes::Units::DEGREES, Axes::Units::QSPACE});
    }
}

std::function<double(double)> AngularReflectometryConverter::getTraslatorTo(Axes::Units units) const
{
    switch (units) {
    case Axes::Units::RADIANS:
        return [](double value) { return value; };
    case Axes::Units::DEGREES:
        return [](double value) { return Units::rad2deg(value); };
    case Axes::Units::QSPACE:
        return [wl = m_wavelength](double value) { return getQ(wl, value); };
    case Axes::Units::RQ4:
        return [wl = m_wavelength](double value) { return getQ(wl, value); };
    default:
        throwUnitsError("AngularReflectometryConverter::getTraslatorTo", availableUnits());
    }
}


//  ************************************************************************************************
//  class WavenumberReflectometryConverter
//  ************************************************************************************************

WavenumberReflectometryConverter::WavenumberReflectometryConverter(const QSpecScan& handler)
    : m_axis(handler.coordinateAxis()->clone())
{
}

WavenumberReflectometryConverter::~WavenumberReflectometryConverter() = default;

WavenumberReflectometryConverter* WavenumberReflectometryConverter::clone() const
{
    return new WavenumberReflectometryConverter(*this);
}

//! Returns the list of all available units
std::vector<Axes::Units> WavenumberReflectometryConverter::availableUnits() const
{
    return {Axes::Units::NBINS, Axes::Units::QSPACE, Axes::Units::RQ4};
}

WavenumberReflectometryConverter::WavenumberReflectometryConverter(
    const WavenumberReflectometryConverter& other)
    : m_axis(std::unique_ptr<const IAxis>(other.coordinateAxis()->clone()))
{
}

//! Creates name map for axis in various units
std::vector<std::map<Axes::Units, std::string>>
WavenumberReflectometryConverter::createNameMaps() const
{
    return {AxisNames::specAxisQ};
}

//! Returns translating functional (inv. nm --> desired units)
std::function<double(double)>
WavenumberReflectometryConverter::getTraslatorTo(Axes::Units units) const
{
    switch (units) {
    case Axes::Units::QSPACE:
        return [](double value) { return value; };
    case Axes::Units::RQ4:
        return [](double value) { return value; };
    default:
        throwUnitsError("WavenumberReflectometryConverter::getTraslatorTo", availableUnits());
    }
}
