//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Unit/ICoordSystem.cpp
//! @brief     Implements ICoordSystem class.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Unit/ICoordSystem.h"
#include "Device/Data/OutputData.h"
#include "Device/Unit/AxisNames.h"


ICoordSystem::~ICoordSystem() = default;

std::unique_ptr<OutputData<double>> ICoordSystem::createOutputData(Axes::Units units) const
{
    std::unique_ptr<OutputData<double>> result = std::make_unique<OutputData<double>>();
    for (size_t i = 0; i < dimension(); ++i)
        result->addAxis(*createConvertedAxis(i, units));
    result->setAllTo(0.0);
    return result;
}

std::string ICoordSystem::axisName(size_t i_axis, const Axes::Units units) const
{
    const auto& name_maps = createNameMaps();
    ASSERT(i_axis < name_maps.size());
    const auto& name_map = name_maps[i_axis];
    const auto& it = name_map.find(units == Axes::Units::UNDEFINED ? defaultUnits() : units);
    ASSERT(it != name_map.cend());
    return it->second;
}

std::unique_ptr<OutputData<double>>
ICoordSystem::createConvertedData(const OutputData<double>& data, Axes::Units units) const
{
    std::unique_ptr<OutputData<double>> result(new OutputData<double>);
    for (size_t i = 0; i < data.rank(); ++i)
        result->addAxis(*createConvertedAxis(i, units));
    result->setRawDataVector(data.getRawDataVector());
    return result;
}

void ICoordSystem::throwUnitsError(std::string method, std::vector<Axes::Units> available) const
{
    std::stringstream ss;
    ss << "Unit type error in " << method
       << ": unknown or unsupported unit type. Available units "
          "are:\n";
    for (auto unit : available)
        ss << axisUnitLabel.at(unit) << "\n";
    throw std::runtime_error(ss.str());
}

Axes::Units ICoordSystem::substituteDefaultUnits(Axes::Units units) const
{
    return units == Axes::Units::UNDEFINED ? defaultUnits() : units;
}
