//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Unit/ICoordSystem.h
//! @brief     Defines interface ICoordSystem and possible axis units.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifdef SWIG
#error no need to expose this header to Swig
#endif

#ifndef USER_API
#ifndef BORNAGAIN_DEVICE_UNIT_IUNITCONVERTER_H
#define BORNAGAIN_DEVICE_UNIT_IUNITCONVERTER_H

#include "Base/Types/ICloneable.h"
#include "Device/Unit/Axes.h"
#include <map>
#include <memory>
#include <string>
#include <vector>

class IAxis;
template <class T> class OutputData;


//! Interface to provide axis translations to different units for simulation output

//! Child classes are currently declared in
//! - Device/Detector/CoordSystem2D.h,
//! - Core/Scan/CoordSystem1D.h.

//! @ingroup simulation_internal

class ICoordSystem : public ICloneable {
public:
    virtual ~ICoordSystem();

    virtual ICoordSystem* clone() const = 0;

    //! Returns zero-valued output data array in specified units
    std::unique_ptr<OutputData<double>> createOutputData(Axes::Units units) const;

    virtual size_t dimension() const = 0;

    virtual double calculateMin(size_t i_axis, Axes::Units units) const = 0;
    virtual double calculateMax(size_t i_axis, Axes::Units units) const = 0;
    virtual size_t axisSize(size_t i_axis) const = 0;

    std::string axisName(size_t i_axis, const Axes::Units units = Axes::Units::UNDEFINED) const;

    virtual std::vector<Axes::Units> availableUnits() const = 0;
    virtual Axes::Units defaultUnits() const = 0;

#ifndef SWIG
    virtual std::unique_ptr<const IAxis> createConvertedAxis(size_t i_axis,
                                                             Axes::Units units) const = 0;

    //! Creates OutputData array in converter units.
    virtual std::unique_ptr<OutputData<double>> createConvertedData(const OutputData<double>& data,
                                                                    Axes::Units units) const;
#endif // SWIG

protected:
    Axes::Units substituteDefaultUnits(Axes::Units units) const;
#ifndef SWIG
    [[noreturn]] void throwUnitsError(std::string method, std::vector<Axes::Units> available) const;
#endif // SWIG

private:
    virtual std::vector<std::map<Axes::Units, std::string>> createNameMaps() const = 0;
};

#endif // BORNAGAIN_DEVICE_UNIT_IUNITCONVERTER_H
#endif // USER_API
